/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-ctype.h"
#include "my-stdio.h"
#include "my-string.h"

#include "config.h"
#include "db.h"
#include "exceptions.h"
#include "list.h"
#include "log.h"
#include "match.h"
#include "ref_count.h"
#include "server.h"
#include "storage.h"
#include "streams.h"
#include "structures.h"
#include "utils.h"

/*
 * These versions of strcasecmp() and strncasecmp() depend on ASCII.
 */

char  cmap[257] = "\
\000\001\002\003\004\005\006\007\010\011\012\013\014\015\016\017\
\020\021\022\023\024\025\026\027\030\031\032\033\034\035\036\037\
\040\041\042\043\044\045\046\047\050\051\052\053\054\055\056\057\
\060\061\062\063\064\065\066\067\070\071\072\073\074\075\076\077\
\100\141\142\143\144\145\146\147\150\151\152\153\154\155\156\157\
\160\161\162\163\164\165\166\167\170\171\172\133\134\135\136\137\
\140\141\142\143\144\145\146\147\150\151\152\153\154\155\156\157\
\160\161\162\163\164\165\166\167\170\171\172\173\174\175\176\177\
\200\201\202\203\204\205\206\207\210\211\212\213\214\215\216\217\
\220\221\222\223\224\225\226\227\230\231\232\233\234\235\236\237\
\240\241\242\243\244\245\246\247\250\251\252\253\254\255\256\257\
\260\261\262\263\264\265\266\267\270\271\272\273\274\275\276\277\
\300\301\302\303\304\305\306\307\310\311\312\313\314\315\316\317\
\320\321\322\323\324\325\326\327\330\331\332\333\334\335\336\337\
\340\341\342\343\344\345\346\347\350\351\352\353\354\355\356\357\
\360\361\362\363\364\365\366\367\370\371\372\373\374\375\376\377";

int
mystrcasecmp(register const char *s, register const char *t)
{
    register char	*c = cmap;

    while (c[*s] == c[*t++]) {
	if (!*s++)
	    return 0;
    }
    return(c[*s] - c[*--t]);
}

int
mystrncasecmp(register const char *s, register const char *t,
		   register int n)
{
    register char	*c = cmap;
    
    if (!n) return 0;
    while (c[*s] == c[*t++]) {
	if (!*s++ || !--n)
	    return 0;
    }
    return(c[*s] - c[*--t]);
}

int
verbcasecmp(const char *verb, const char *word)
{
    const char		       *w;
    enum { none, inner, end }	star;
    char		       *c = cmap;

    while (*verb) {
	w = word;
	star = none;
	while (1) {
	    while (*verb == '*') {
		verb++;
		star = (!*verb || *verb == ' ') ? end : inner;
	    }
	    if (!*verb || *verb == ' ' || !*w || c[*w] != c[*verb])
		break;
	    w++;
	    verb++;
	}
	if (!*w ? (star != none || !*verb || *verb == ' ')
	        : (star == end))
	    return 1;
	while (*verb && *verb != ' ')
	    verb++;
	while (*verb == ' ')
	    verb++;
    }
    return 0;
}

void
object_match_failed(Objid  player, Objid thing, const char *s)
{
    Stream     *str = new_stream(100);

    switch (thing) {
      case FAILED_MATCH:
	stream_printf(str, "I don't see \"%s\" here.", s);
	break;
      case AMBIGUOUS:
	stream_printf(str, "\"%s\":  I don't know which one you mean.", s);
	break;
      default:
	stream_printf(str, "\"%s\":  Invalid object.", s);
	break;
    }
    notify(player, stream_contents(str));
    free_stream(str);
}

void
free_var(Var v)
{
    int		i;

    switch ((int) v.type) {
      case TYPE_STR:
        if (v.v.str)
            free_str(v.v.str);
	break;
      case TYPE_LIST:
	if (delref(v.v.list) == 0) {
	    for(i = 1; i <= v.v.list[0].v.num; i++)
		free_var( v.v.list[i]);
	    myfree((void *) v.v.list, M_LIST);
	}
	break;
      case TYPE_NUM:
      case TYPE_OBJ:
      case TYPE_ERR:
      case TYPE_CLEAR:
      case (int) TYPE_ANY:
	break;
      default:
	panic("FREE_VAR: Impossible var type.\n");
	break;
    }
}

Var
var_ref(Var v)
{
    switch((int) v.type) {
      case TYPE_STR:
	addref(v.v.str);
	break;
      case TYPE_LIST:
	addref(v.v.list);
	break;
      case TYPE_NUM:
      case TYPE_OBJ:
      case TYPE_ERR:
      case TYPE_CLEAR:
      case (int) TYPE_ANY:
	break;
      default:
	errlogf("VAR_REF: Impossible var type: %d\n", (int)v.type);
    }
    return v;
}

Var
var_dup(Var v)
{
    int		i;
    Var		newlist;
    
    switch ((int) v.type) {
      case TYPE_STR:
	v.v.str = str_dup(v.v.str);
	break;
      case TYPE_LIST:
	newlist = new_list(v.v.list[0].v.num);
	for(i = 1; i <= v.v.list[0].v.num; i++) {
	    newlist.v.list[i] = var_ref(v.v.list[i]);
	}
	v = newlist;
	break;
      case TYPE_NUM:
      case TYPE_OBJ:
      case TYPE_ERR:
      case (int) TYPE_ANY:
	break;
      default:
	errlogf("VAR_DUP: Impossible var type: %d\n", (int)v.type);
    }
    return v;
}

int
compare_ints(int a, int b)
{
    if (a < b)
	return -1;
    else if (a == b)
	return 0;
    else
	return 1;
}

int
is_true(Var v)
{
    return (   (v.type == TYPE_NUM  &&  v.v.num != 0)
	    || (v.type == TYPE_STR  &&  v.v.str  &&  *v.v.str != '\0')
	    || (v.type == TYPE_LIST &&  v.v.list[0].v.num != 0));
}

int
equality(Var lhs, Var rhs, int case_matters)
{
    if (lhs.type != rhs.type)
	return 0;
    else {
	switch (lhs.type) {
	  case TYPE_NUM:
	    return lhs.v.num == rhs.v.num;
	  case TYPE_OBJ:
	    return lhs.v.obj == rhs.v.obj;
	  case TYPE_ERR:
	    return lhs.v.err == rhs.v.err;
	  case TYPE_STR:
	    if (case_matters)
		return !strcmp(lhs.v.str, rhs.v.str);
	    else
		return !mystrcasecmp(lhs.v.str, rhs.v.str);
	  case TYPE_LIST:
	    if (lhs.v.list[0].v.num != rhs.v.list[0].v.num)
		return 0;
	    else {
		int	i;

		for (i = 1; i <= lhs.v.list[0].v.num; i++) {
		    if (!equality(lhs.v.list[i], rhs.v.list[i], case_matters))
			return 0;
		}
		return 1;
	    }
	  default:
	    errlogf("EQUALITY: Impossible var type: %d\n", (int)lhs.type);
	}
    }
    return 0;
}

void
init_pi(Parse_Info *pi)
{
    pi->player = pi->dobj = pi->iobj = pi->caller 
 	   = pi->this = pi->vloc = NOTHING;

    pi->verb = pi->verbname = pi->dobjstr = pi->iobjstr = pi->argstr
    = pi->prepstr = 0;

    pi->prep = NOTHING;

    pi->args.type = TYPE_NUM;
    pi->args.v.list = 0;
}

void
free_pi(Parse_Info *pi, int free_pi_itself)
{
    if (pi == 0)
	return;
    if (pi->args.v.list)
	free_var(pi->args);
    if (pi->argstr)
	free_str(pi->argstr);
    if (pi->dobjstr)
	free_str(pi->dobjstr);
    if (pi->iobjstr)
	free_str(pi->iobjstr);
    if (pi->prepstr)
	free_str(pi->prepstr);
    if (pi->verb)
	free_str(pi->verb);
    if (pi->verbname)
	free_str(pi->verbname);
    if (free_pi_itself)
	myfree((void *) pi, M_PI);
}

Parse_Info *
new_pi(void)
{
    Parse_Info   *new = (Parse_Info *) mymalloc(sizeof(Parse_Info), M_PI);

    init_pi(new);

    return new;
}

void
write_activ_as_pi(activation a)
{
    Var dummy;
    dummy.type = TYPE_NUM;
    dummy.v.num = -111;

    write_var(dummy);
    db_printf("%d %d %d %d %d %d %d %d %d\n",
	      a.this, -7, -8, a.player, -9,
	      a.progr, a.vloc, -10, a.debug);
    write_string("No");
    write_string("More");
    write_string("Parse");
    write_string("Infos");
    write_string(a.verb);
    write_string(a.verbname);
}

int
read_activ_as_pi(FILE *f, activation *a)
{
    Var dummy;
    Objid o1;
    int i1;
    
    dummy = read_var(f); free_var(dummy);
    if (fscanf(f, "%d %d %d %d %d %d %d %d %d",
	       &(a->this), &o1, &o1, &a->player, &o1,
	       &a->progr, &a->vloc, &i1, &a->debug) != 9
	|| fgetc(f) != '\n') {
	errlogf("READ_A: Bad numbers.\n");
	return -1;
    }
    read_string(f); /* was argstr */
    read_string(f); /* was dobjstr */
    read_string(f); /* was iobjstr */
    read_string(f); /* was prepstr */
    a->verb = str_dup(read_string(f));
    a->verbname = str_dup(read_string(f));
    return 0;
}

char *
strsub(const char *source, const char *what, const char *with, int case_counts)
{
    static Stream *str = 0;
    int		lwhat = strlen(what);

    if (str == 0)
	str = new_stream(100);

    while (*source) {
	if (! (case_counts ? strncmp(source, what, lwhat)
	                   : mystrncasecmp(source, what, lwhat))) {
	    stream_add_string(str, with);
	    source += lwhat;
	} else
	    stream_add_char(str, *source++);
    }

    return reset_stream(str);
}

int
strindex(const char *source, const char *what, int case_counts)
{
    const char	*s, *e;
    int		lwhat = strlen(what);

    for (s = source, e = source + strlen(source) - lwhat; s <= e; s++) {
	if (! (case_counts ? strncmp(s, what, lwhat)
	                   : mystrncasecmp(s, what, lwhat))) {
	    return s - source + 1;
	}
    }
    return 0;
}

int
strrindex(const char *source, const char *what, int case_counts)
{
    const char	*s;
    int		lwhat = strlen(what);

    for (s = source + strlen(source) - lwhat; s >= source; s--) {
	if (! (case_counts ? strncmp(s, what, lwhat)
	                   : mystrncasecmp(s, what, lwhat))) {
	    return s - source + 1;
	}
    }
    return 0;
}

Var
get_system_property(const char *name)
{
    Var		value;
    enum error	e;

   if (!valid(SYSTEM_OBJECT)) {
	value.type = TYPE_ERR;
	value.v.err = E_INVIND;
	return value;
    }
    e = get_non_bi_prop(SYSTEM_OBJECT, name, &value, 0);
    if (e != E_NONE) {
	value.type = TYPE_ERR;
	value.v.err = e;
    }
    return value;
}

Objid
get_system_object(const char *name)
{
    Var		value;

    value = get_system_property(name);
    if (value.type != TYPE_OBJ)
	return NOTHING;
    else
	return value.v.obj;
}

char rcsid_utils[] = "$Id: utils.c,v 1.13 1992/10/23 23:03:47 pavel Exp $";

/* $Log: utils.c,v $
 * Revision 1.13  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.12  1992/10/23  22:23:18  pavel
 * Eliminated all uses of the useless macro NULL.
 *
 * Revision 1.11  1992/10/17  20:57:08  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 * Removed useless #ifdef of STRCASE.
 *
 * Revision 1.10  1992/09/14  17:39:53  pjames
 * Moved db_modification code to db modules.
 *
 * Revision 1.9  1992/09/08  21:55:47  pjames
 * Updated #includes.
 *
 * Revision 1.8  1992/09/03  16:23:49  pjames
 * Make cmap[] visible.  Added TYPE_CLEAR handling.
 *
 * Revision 1.7  1992/08/28  16:23:41  pjames
 * Changed vardup to varref.
 * Changed myfree(*, M_STRING) to free_str(*).
 * Added `varref()'.
 * Changed `free_var()' to check `delref()' before freeing.
 * Removed `copy_pi()'.
 *
 * Revision 1.6  1992/08/18  00:54:40  pavel
 * Fixed typo.
 *
 * Revision 1.5  1992/08/18  00:41:14  pavel
 * Fixed boundary-condition bugs in index() and rindex(), when the search string
 * is empty.
 *
 * Revision 1.4  1992/08/11  17:26:56  pjames
 * Removed read/write Parse_Info procedures.
 *
 * Revision 1.3  1992/08/10  16:40:20  pjames
 * Added is_true (from execute.c) and *_activ_as_pi routines.  Added a
 * check for null parse_info in write_pi, because parse_infos are no
 * longer stored in each activation.
 *
 * Revision 1.2  1992/07/21  00:07:48  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
